#include <algorithm>
#include <cassert>
#include <iostream>
#include <memory>
#include "Constants.h"
#include "Random378.h"
#include "Submarine.h"


class SubmarineInstance {
public:
    SubmarineInstance(const Submarine* prototype) {
        _prototype = prototype;
        _automata.reset(prototype->_behavior.instance());

        // Set the radius with a gaussian at some point.
        const float theta = frandMT() * PI_2;
        const float radius = frandMT() * maxStartingRadius;

        _posX = radius * sin(theta);
        _posY = radius * cos(theta);

        _direction = frandMT() * PI_2;
        //_turnRate = 0;
        _speed = 3;

        // Distribute weapon units.
        _fore = _aft = _pulses = 0;
        for (size_t i = 0; i < totalWeapons; ++i) {
            switch (prototype->_weapons[i]) {
                case 'F': ++_fore;   break;
                case 'A': ++_aft;    break;
                case 'P': ++_pulses; break;
            }
        }

        /*
        const float* weapons = prototype->_weapons;
        float sum = weapons[0] + weapons[1] + weapons[2];
        if (sum == 0) {
            _fore = int(totalWeapons / 3);
            _aft  = int(totalWeapons / 3);
        } else {
            _fore = int(totalWeapons * weapons[0] / sum);
            _aft  = int(totalWeapons * weapons[1] / sum);
        }
        _pulses = totalWeapons - _fore - _aft;
        */

        _damage = 0;

        std::fill(_actionCounts, _actionCounts + actionCount, 0);
        std::fill(_shots, _shots + weaponCount, 0);
        std::fill(_hits, _hits + weaponCount, 0);
    }

    float getX() const {
        return _posX;
    }

    float getY() const {
        return _posY;
    }

    float getDirection() const {
        return _direction;
    }

    const float getDamage() const {
        return _damage;
    }

    void hitFore()  { ++_hits[0]; }
    void hitAft()   { ++_hits[1]; }
    void hitPulse() { ++_hits[2]; }

    void damage(float amount) {
        _damage += amount;
    }

    bool isDead() const {
        return _damage > maxDamage;
    }
    
    float getAccuracy() const {
        float a0 = float(_hits[0]) / _shots[0];
        float a1 = float(_hits[1]) / _shots[1];
        float a2 = float(_hits[2]) / _shots[2];
        return a0 + a1 + a2;
    }

    const int* getActionCounts() const {
        return _actionCounts;
    }

    const int* getShotCounts() const {
        return _shots;
    }

    const int* getHitCounts() const {
        return _hits;
    }

    int initialAction() {
        return perform(_automata->initialAction());
    }

    void updatePosition(float dt) {
        _posX += sin(_direction) * _speed * dt;
        _posY += cos(_direction) * _speed * dt;
        //_direction += _turnRate / 10.0f * dt;
    }

    int updateAction(int distance, int angle) {
        int inputs[inputCount];
        inputs[ForwardCount] = _fore;
        inputs[AftCount]     = _aft;
        inputs[PulseCount]   = _pulses;
        inputs[Distance]     = distance;
        inputs[Angle]        = angle;
        return perform(_automata->step(inputs));
    }

private:
    int perform(int action) {
        ++_actionCounts[action];

        switch (action) {
            case Think:
                break;

            case FireForward:
                if (_fore > 0) {
                    --_fore;
                    ++_shots[0];
                } else {
                    action = Think;
                }
                break;

            case FireAft:
                if (_aft > 0) {
                    --_aft;
                    ++_shots[1];
                } else {
                    action = Think;
                }
                break;

            case FirePulse:
                if (_pulses > 0) {
                    --_pulses;
                    ++_shots[2];
                } else {
                    action = Think;
                }
                break;

            case TurnLeft:
                _direction -= PI / 8;
                //_turnRate = std::max(-2, _turnRate - 1);
                break;

            case TurnRight:
                _direction += PI / 8;
                //_turnRate = std::min(2, _turnRate + 1);
                break;

            case SpeedUp:
                _speed = std::min(5, _speed + 1);
                break;

            case SlowDown:
                _speed = std::max(1, _speed - 1);
                break;

            default:
                assert(false);
        }
        return action;
    }

    const Submarine* _prototype;
    std::auto_ptr<GPAutomata::Instance> _automata;
    float _posX;
    float _posY;
    float _direction;

    //int _turnRate;  // [-2, 2], defaults to 0
    int _speed;     // [1, 5], defaults to 3

    int _fore;
    int _aft;
    int _pulses;

    float _damage;  // 0

    // Statistics.
    int _actionCounts[actionCount];
    int _shots[weaponCount];
    int _hits[weaponCount];
};


Submarine::Submarine() {
    std::fill(_weapons, _weapons + totalWeapons, 'F');
}


char randomWeapon() {
    switch (lrandMT() % 3) {
        case 0:  return 'F';
        case 1:  return 'A';
        default: return 'P';
    }
}


void Submarine::randomize() {
    std::generate(_weapons, _weapons + totalWeapons, randomWeapon);
    _behavior.randomize();
}


template<typename T>
T sqr(T t) {
    return t * t;
}


float distance(const SubmarineInstance& one, const SubmarineInstance& two) {
    return sqrt(sqr(one.getX() - two.getX()) + sqr(one.getY() - two.getY()));
}


float angularDistance(float d1, float d2) {
    float x1 = sin(d1);
    float y1 = cos(d1);
    float x2 = sin(d2);
    float y2 = cos(d2);
    return acos(x1 * x2 + y1 * y2);
    //return fabs(fmod(d1 - d2, 2 * PI));
}

float angularDistance2(float d1, float d2) {
    float d = d2 - d1;
    while (d > PI) {
        d -= PI_2;
    }
    while (d < -PI) {
        d += PI_2;
    }
    return d;
}


void executeAction(int action, SubmarineInstance& performer, SubmarineInstance& other) {
    float direction = performer.getDirection();
    float dx = other.getX() - performer.getX();
    float dy = other.getY() - performer.getY();
    float dist = distance(performer, other);
    float atOther = atan2(dx, dy);

    switch (action) {
        case FireForward:
            if (dist < 100.0f && angularDistance(direction, atOther) < PI / 8) {
                performer.hitFore();
                other.damage(forwardTorpedoDamage);
            }
            break;

        case FireAft:
            if (dist < 100.0f && angularDistance(direction + PI, atOther) < PI / 8) {
                performer.hitAft();
                other.damage(aftTorpedoDamage);
            }
            break;

        case FirePulse:
            if (dist < pulseRange) {
                performer.hitPulse();
                other.damage(pulseDamage / (1 + sqr(dist / 5.0f)));
            }
            break;
    }
}


BattleResult makeResult(float time, Statistics distance, const SubmarineInstance& one, const SubmarineInstance& two) {
    BattleResult rv;

    rv.time = time;
    rv.distance = distance;

    if (one.isDead() && two.isDead()) {
        rv.winner = BattleResult::TIE;
    } else if (one.isDead()) {
        rv.winner = BattleResult::TWO;
    } else if (two.isDead()) {
        rv.winner = BattleResult::ONE;

    } else if (one.getAccuracy() > two.getAccuracy()) {
        rv.winner = BattleResult::ONE;
    } else if (one.getAccuracy() < two.getAccuracy()) {
        rv.winner = BattleResult::TWO;

//    } else if (one.getDamage() > two.getDamage()) {
//        rv.winner = BattleResult::TWO;
//    } else if (one.getDamage() < two.getDamage()) {
//        rv.winner = BattleResult::ONE;
    } else {
        rv.winner = BattleResult::TIE;
    }

    const int* ac1 = one.getActionCounts();
    const int* ac2 = two.getActionCounts();

    std::fill(rv.actionCounts, rv.actionCounts + actionCount, 0);
    for (size_t i = 0; i < actionCount; ++i) {
        rv.actionCounts[i] += ac1[i];
        rv.actionCounts[i] += ac2[i];
    }

    std::fill(rv.shotCounts, rv.shotCounts + weaponCount, 0);
    std::fill(rv.hitCounts, rv.hitCounts + weaponCount, 0);
    for (size_t i = 0; i < weaponCount; ++i) {
        rv.shotCounts[i] += one.getShotCounts()[i];
        rv.shotCounts[i] += two.getShotCounts()[i];
        rv.hitCounts[i] += one.getHitCounts()[i];
        rv.hitCounts[i] += two.getHitCounts()[i];
    }

    return rv;
}


BattleResult Submarine::battle(Submarine& one, Submarine& two) {
    std::auto_ptr<SubmarineInstance> s1(one.instance());
    std::auto_ptr<SubmarineInstance> s2(two.instance());

    int action1 = s1->initialAction();
    int action2 = s2->initialAction();

    Statistics distanceStatistics;

    const float dt = battleDT;
    for (float time = 0; time < battleTimeLimit; time += dt) {
        executeAction(action1, *s1, *s2);
        executeAction(action2, *s2, *s1);

        if (s1->isDead() || s2->isDead()) {
            return makeResult(time, distanceStatistics, *s1, *s2);
        }

        s1->updatePosition(dt);
        s2->updatePosition(dt);

        float d = distance(*s1, *s2);
        distanceStatistics.add(d);

        float s1x = s1->getX();
        float s1y = s1->getY();
        float s2x = s2->getX();
        float s2y = s2->getY();

        float angle1 = atan2(s2x - s1x, s2y - s1y);
        float angle2 = atan2(s1x - s2x, s1y - s2y);

        action1 = s1->updateAction(int(d), int(angularDistance2(s1->getDirection(), angle1) * 180 / PI));
        action2 = s2->updateAction(int(d), int(angularDistance2(s2->getDirection(), angle2) * 180 / PI));
    }

    // Battle has timed out.
    return makeResult(battleTimeLimit, distanceStatistics, *s1, *s2);
}


void Submarine::crossover(Submarine& one, Submarine& two) {
    if (lrandMT()) {
        int s1 = lrandMT() % totalWeapons;
        int s2 = lrandMT() % totalWeapons;
        for (int i = s1; i != s2; i = (i + 1) % totalWeapons) {
            std::swap(one._weapons[i], two._weapons[i]);
        }
    } else {
        ::crossover(one._behavior, two._behavior);
    }
}


void Submarine::mutate() {
    if (randomBit()) {
        _weapons[lrandMT() % totalWeapons] = randomWeapon();
    } else {
        _behavior.mutate();
    }
}


SubmarineInstance* Submarine::instance() const {
    return new SubmarineInstance(this);
}
